"""Generate Color Palettes w.r.t. input names. 
"""

import colorsys
import matplotlib
from STHD import config


def _divide_hue(hue_range, n, hue_shift):
    hue_delta = (hue_range[1] - hue_range[0]) / (n - 1)
    res = []
    for i in range(n):
        res.append(hue_range[0] + i * hue_delta + hue_shift)
    return res


def _construct_color(hue_list, saturation, lightness, color_format):
    res = []
    for hue in hue_list:
        r, g, b = colorsys.hls_to_rgb(hue, saturation, lightness)
        if color_format == "hex":
            color_code = "#{:02x}{:02x}{:02x}".format(
                int(r * 255), int(g * 255), int(b * 255)
            )
        elif color_format == "rgb":
            color_code = (r, g, b)
        res.append(color_code)
    return res


def _generate_color_palette(
    names,
    hue_range,
    saturation,
    lightness,
    hue_shift=0,
    color_format="rgb",  # either 'rgb' or 'hex'
):
    hues = _divide_hue(hue_range, len(names), hue_shift)
    colors = _construct_color(hues, saturation, lightness, color_format)

    res = dict(zip(names, colors))
    return res


def get_color_map_1(genemeanpd_filtered):
    # define colormap
    names = ["p_ct_" + i for i in list(genemeanpd_filtered.columns)]
    tumor_colors = _generate_color_palette(
        names=names[:11],
        hue_range=[0, 0.2],
        saturation=0.8,
        lightness=1,
        hue_shift=0,
        color_format="hex",
    )
    normal_colors = _generate_color_palette(
        names=names[11:],
        hue_range=[0.25, 0.8],
        saturation=0.7,
        lightness=0.7,
        hue_shift=0,
        color_format="hex",
    )
    cmap = {**tumor_colors, **normal_colors}
    cmap["ambiguous"] = [0.65] * 3
    return cmap


def prepare_palette(cmap, adata, ctcol="STHD_pred_ct"):
    data_cmap = []
    for p in sorted(list(set(adata.obs[ctcol]))):
        cur_color = cmap[p]
        data_cmap.append(cur_color)
    palette = matplotlib.colors.ListedColormap(data_cmap)
    return palette


def get_config_colormap(name="colormap_coloncatlas_98"):
    cmap = None

    if name == "colormap_coloncatlas_98":
        cmap = config.colormap_coloncatlas_98
    elif name == "colormap_coloncatlas_98_light":
        cmap = config.colormap_coloncatlas_98_light
    elif name == "colormap_coloncatlas_98_dark":
        cmap = config.colormap_coloncatlas_98_dark
    elif name == "colormap_crc98_ct_group":
        cmap = config.colormap_crc98_ct_group
    return cmap


def adjust_lightness(hex_color, adjustment):
    """Adjust the lightness of a hex color.

    Args:
    ----
        hex_color (str): The hex color code (e.g., '#RRGGBB').
        adjustment (float): The amount by which to adjust the lightness.
                            Positive values make the color lighter, negative values make it darker.
                            Should be between -1 and 1.

    Returns:
    -------
        str: The adjusted hex color code.

    """
    # Remove '#' from the hex color string
    hex_color = hex_color.lstrip("#")

    # Convert the hex color to RGB values
    red = int(hex_color[0:2], 16)
    green = int(hex_color[2:4], 16)
    blue = int(hex_color[4:6], 16)

    # Adjust the lightness
    red = min(255, max(0, int(red + (255 - red) * adjustment)))
    green = min(255, max(0, int(green + (255 - green) * adjustment)))
    blue = min(255, max(0, int(blue + (255 - blue) * adjustment)))

    # Convert the adjusted RGB values back to hex
    adjusted_hex_color = "#{0:02x}{1:02x}{2:02x}".format(red, green, blue)

    return adjusted_hex_color



"""Function for region of interest analysis"""

import pandas as pd


def barcode_convert(sthdata, nspot=4, switch_index=False):
    """Convert barcodes from 2um bin to 8um bin, add to .adata.obs

    Args:
    ----
    sthdata: Data object containing 'adata'.
    nspot: number of spots to bin together.
    switch_index: whether to set index to converted barcodes.

    Returns:
    -------
    Modified 'sthdata' in place by updating the 8um barcodes as index.
    add the old 2um barcodes into the sthdata.adata.obs['barcodes_2um']

    """
    # process row and column num for 8um bins
    # sthdata = sthd_data.copy()
    array_row_8um = (
        pd.Series(sthdata.adata.obs.index)
        .str.split("-")
        .str.get(0)
        .str.split("_")
        .str.get(2)
        .astype(int)
        .apply(lambda x: str(x // nspot).zfill(5))
    )
    array_col_8um = (
        pd.Series(sthdata.adata.obs.index)
        .str.split("-")
        .str.get(0)
        .str.split("_")
        .str.get(3)
        .astype(int)
        .apply(lambda x: str(x // nspot).zfill(5))
    )
    # Update the barcode index in 'sthdata' to 8um bins
    sthdata.adata.obs["barcodes_2um"] = sthdata.adata.obs.index
    sthdata.adata.obs["barcodes_8um"] = (
        "s_008um_" + array_row_8um + "_" + array_col_8um + "-1"
    )
    if switch_index:
        sthdata.adata.obs.set_index("barcodes_8um")

    return sthdata



"""
patch_path = '../testdata/crop10'
refile = '../testdata/crc_average_expr_genenorm_lambda_98ct_4618gs.txt'
sthdata = train.load_data_with_pdata(patch_path)

binadata = binning_fast.get_sthd_guided_bin_adata(sthdata, nspot= 4, use_csr = True)

"""

import argparse
import os
import sys
from collections import defaultdict

import anndata
import numpy as np
import pandas as pd
import scipy.sparse as sp
from tqdm import tqdm
from STHD import train


def get_raw_bins(adata, nspot):
    """Binning data according to array_row, array_col.
    The bin_id is "{bin_row}||{bin_col}"
    """
    array_row_min, array_col_min = (
        adata.obs["array_row"].min(),
        adata.obs["array_col"].min(),
    )
    bin_row = (
        ((adata.obs["array_row"].values - array_row_min) // nspot).astype(int) * nspot
        + array_row_min
    ).astype(int)
    bin_col = (
        ((adata.obs["array_col"].values - array_col_min) // nspot).astype(int) * nspot
        + array_col_min
    ).astype(int)

    bins = defaultdict(
        list
    )  # []defaultdict must guarantee order of bins. e.g. python>3.7
    for i in tqdm(range(len(bin_row))):
        bins[f"{bin_row[i]}||{bin_col[i]}"].append(i)
    return bins


def get_bins(adata, nspot):
    """Binning data according to array_row, array_col, and cell type predictions.
    The bin_id is "{bin_row}||{bin_col}||{cell_type}"
    """
    array_row_min, array_col_min = (
        adata.obs["array_row"].min(),
        adata.obs["array_col"].min(),
    )
    bin_row = (
        ((adata.obs["array_row"].values - array_row_min) // nspot).astype(int) * nspot
        + array_row_min
    ).astype(int)
    bin_col = (
        ((adata.obs["array_col"].values - array_col_min) // nspot).astype(int) * nspot
        + array_col_min
    ).astype(int)
    bin_ct = adata.obs["STHD_pred_ct"].values

    bins = defaultdict(list)
    for i in tqdm(range(len(bin_row))):
        bins[f"{bin_row[i]}||{bin_col[i]}||{bin_ct[i]}"].append(i)
    return bins


def bin_X(adata, bins):
    """Use data in adata.X."""
    bin_value = []
    for bid in tqdm(bins, total=len(bins)):
        cur_ids = bins[bid]
        cur_value = np.squeeze(np.asarray(adata.X[cur_ids].mean(axis=0)))
        bin_value.append(cur_value)
    return np.array(bin_value)


def bin_obsm(adata, bins, obsm_key_list=["spatial"]):
    obsm_res = dict()
    print("[Log] binning obsm...")
    for obsm_key in obsm_key_list:
        bin_value = []
        for bid in tqdm(bins, total=len(bins)):
            cur_ids = bins[bid]
            cur_value = adata.obsm[obsm_key][cur_ids].mean(axis=0)
            bin_value.append(cur_value)
        obsm_res[obsm_key] = np.array(bin_value)
    return obsm_res


def bin_raw_obs(adata, bins):
    val_bin_row = []
    val_bin_col = []
    val_barcodes_included = []
    val_n_counts = []
    val_bin_n_spot = []

    barcodes = np.array(list(adata.obs.index))

    bin_id_list = list(bins.keys())
    print("[Log] binning obs...")
    for bid in tqdm(bin_id_list, total=len(bin_id_list)):
        cur_bin_row, cur_bin_col = bid.split("||")
        cur_ids = bins[bid]

        cur_bin_n_spot = len(cur_ids)
        cur_barcodes_included = list(barcodes[cur_ids])
        cur_n_counts = adata.X[cur_ids].sum()

        val_bin_row.append(cur_bin_row)
        val_bin_col.append(cur_bin_col)
        val_barcodes_included.append(cur_barcodes_included)
        val_n_counts.append(cur_n_counts)
        val_bin_n_spot.append(cur_bin_n_spot)
    return pd.DataFrame(
        {
            "bin_row": val_bin_row,
            "bin_col": val_bin_col,
            "barcodes_included": val_barcodes_included,
            "n_counts": val_n_counts,
            "bin_n_spot": val_bin_n_spot,
        },
        index=bin_id_list,
    )


def bin_obs(adata, bins):
    val_bin_row = []
    val_bin_col = []
    val_STHD_pred_ct = []
    val_barcodes_included = []
    val_n_counts = []
    val_bin_n_spot = []

    barcodes = np.array(list(adata.obs.index))

    bin_id_list = list(bins.keys())
    print("[Log] binning obs...")
    for bid in tqdm(bin_id_list, total=len(bin_id_list)):
        cur_bin_row, cur_bin_col, cur_STHD_pred_ct = bid.split("||")
        cur_ids = bins[bid]

        cur_bin_n_spot = len(cur_ids)
        cur_barcodes_included = list(barcodes[cur_ids])
        cur_n_counts = adata.X[cur_ids].sum()

        val_bin_row.append(cur_bin_row)
        val_bin_col.append(cur_bin_col)
        val_STHD_pred_ct.append(cur_STHD_pred_ct)
        val_barcodes_included.append(cur_barcodes_included)
        val_n_counts.append(cur_n_counts)
        val_bin_n_spot.append(cur_bin_n_spot)
    return pd.DataFrame(
        {
            "bin_row": val_bin_row,
            "bin_col": val_bin_col,
            "STHD_pred_ct": val_STHD_pred_ct,
            "barcodes_included": val_barcodes_included,
            "n_counts": val_n_counts,
            "bin_n_spot": val_bin_n_spot,
        },
        index=bin_id_list,
    )


def bin_obs_v2(adata, bins, extra_obs_col=""):
    """Temp for : extra obs cols, has to be numerical"""
    val_bin_row = []
    val_bin_col = []
    val_STHD_pred_ct = []
    val_barcodes_included = []
    val_n_counts = []
    val_bin_n_spot = []
    val_extra_obs_col = []

    barcodes = np.array(list(adata.obs.index))

    bin_id_list = list(bins.keys())
    print("[Log] binning obs...")
    for bid in tqdm(bin_id_list, total=len(bin_id_list)):
        cur_bin_row, cur_bin_col, cur_STHD_pred_ct = bid.split("||")
        cur_ids = bins[bid]

        cur_bin_n_spot = len(cur_ids)
        cur_barcodes_included = list(barcodes[cur_ids])
        cur_n_counts = adata.X[cur_ids].sum()

        cur_col_value = adata.obs.iloc[cur_ids][extra_obs_col].mean()

        val_bin_row.append(cur_bin_row)
        val_bin_col.append(cur_bin_col)
        val_STHD_pred_ct.append(cur_STHD_pred_ct)
        val_barcodes_included.append(cur_barcodes_included)
        val_n_counts.append(cur_n_counts)
        val_bin_n_spot.append(cur_bin_n_spot)
        val_extra_obs_col.append(cur_col_value)

    return pd.DataFrame(
        {
            "bin_row": val_bin_row,
            "bin_col": val_bin_col,
            "STHD_pred_ct": val_STHD_pred_ct,
            "barcodes_included": val_barcodes_included,
            "n_counts": val_n_counts,
            "bin_n_spot": val_bin_n_spot,
            "extra_obs_col": val_extra_obs_col,
        },
        index=bin_id_list,
    )


def bin_X_csr(adata, bins):
    """Use data in adata.X."""
    bin_values = []
    print("[LOG] Calculating each bin's sum values")
    for bin_name in tqdm(bins, total=len(bins)):
        bin_values.append(sum_csr(bins[bin_name], adata))
    res_csr = concatenate_csr_matrices(bin_values)
    return res_csr


def average_csr(cur_ids, adata):
    cur_value = np.asarray(adata.X[cur_ids].mean(axis=0))
    cur_value = sp.csr_array(cur_value)
    return cur_value


def sum_csr(cur_ids, adata):
    cur_value = np.asarray(adata.X[cur_ids].sum(axis=0))
    cur_value = sp.csr_array(cur_value)
    return cur_value


def concatenate_csr_matrices(matrix_list):
    ncol = matrix_list[0].shape[1]
    for m in matrix_list:
        if m.shape[1] != ncol:
            raise ValueError("Number of columns must be the same for all matrices.")

    n_val = 0
    n_row = 0
    for m in matrix_list:
        n_val += len(m.data)
        n_row += len(m.indptr) - 1

    res_data = np.zeros(n_val)
    res_indices = np.zeros(n_val)
    res_indptr = np.zeros(n_row + 1)

    i = 0  # index for both data and indices
    j = 0  # index for indptr
    s = 0  # adjustment for indptr
    print("[LOG] Constructing adata.X as a sparse matrix")
    for m in tqdm(matrix_list, total=len(matrix_list)):
        res_data[i : i + len(m.data)] = m.data
        res_indices[i : i + len(m.indices)] = m.indices
        res_indptr[j : j + len(m.indptr) - 1] = m.indptr[1:] + s

        i += len(m.data)
        j += len(m.indptr) - 1
        s += m.indptr[-1]

    res_indptr[1:] = res_indptr[:-1]
    res_indptr[0] = 0
    concatenated_matrix = sp.csr_matrix(
        (res_data, res_indices, res_indptr), shape=(len(res_indptr) - 1, m.shape[1])
    )
    return concatenated_matrix


def get_raw_bin_adata(sthdata, nspot=4, min_nspot_to_aggregate=2, use_csr=True):
    """Bin aggregating data, guided by STHD predicted cell type identities of spots. E.g. if one bin has 2 types of spots, they will be aggregated separately, and location is determined by taking mean of x or y for the included spot barcodes
    For full side, must use use_csr. takes 1.5hr for 8million spots -> 4x4 cell type specific binning. ~3GB result
    []todo: barcode string reformat.
    #full_binadata.obs['barcodes_included'] = #full_binadata.obs['barcodes_included'].str.join(sep=',')
    full_binadata.write('../testdata/crop10/full_STHD_guided_binadata.h5ad')
    Params
    ----------
    sthdata:
        STHD class, including adata with spatial, coodinate, fullregimg
    nspot: int
        Number of  spots to square-bin together, along x or along y.
    min_nspot_to_aggregate:
        require that at least min_nspot_to_aggregate number of spots are in the bin to get an aggregated count.
    use_csr:
        Whether to use csr format to save the binned adata.X object. Default to be True.
    """
    bins = get_raw_bins(sthdata.adata, nspot)
    if use_csr:
        X_binned = bin_X_csr(sthdata.adata, bins)
    else:
        X_binned = bin_X(sthdata.adata, bins)
    obs_binned = bin_raw_obs(sthdata.adata, bins)
    obsm_binned = bin_obsm(sthdata.adata, bins)
    binvar = sthdata.adata.var.copy()

    ###########  create binadata ##########
    binadata = anndata.AnnData(
        X_binned,  # raw bin count
        var=binvar,  # gene var
        obs=obs_binned,  # new bin, barcode is array_row||array_col
        obsm=obsm_binned,  # original obsm plus spatial , mean of bin barcode
        uns=sthdata.adata.uns.copy(),  # original uns, image
    )
    # filter minimum spot number
    binadata = binadata[binadata.obs["bin_n_spot"] >= min_nspot_to_aggregate]

    # change list to comma
    binadata.obs["barcodes_included"] = binadata.obs["barcodes_included"].str.join(
        sep=","
    )
    print(binadata)
    return binadata


def get_sthd_guided_bin_adata(
    sthdata,
    pred_col="STHD_pred_ct",
    nspot=4,
    remove_ambiguous_spot=True,
    ambiguous_spot_celltypes=["ambiguous", "filtered"],
    min_nspot_to_aggregate=2,
    use_csr=True,
):
    """Bin aggregating data, guided by STHD predicted cell type identities of spots. E.g. if one bin has 2 types of spots, they will be aggregated separately, and location is determined by taking mean of x or y for the included spot barcodes
    For full side, must use use_csr. takes 1.5hr for 8million spots -> 4x4 cell type specific binning. ~3GB result
    []todo: barcode string reformat.
    #full_binadata.obs['barcodes_included'] = full_binadata.obs['barcodes_included'].str.join(sep=',')
    full_binadata.write('../testdata/crop10/full_STHD_guided_binadata.h5ad')
    Params
    ----------
    sthdata:
        STHD class, including adata with spatial, coodinate, fullregimg
    pred_col:
        column in STHD.obs marking STHD predicted cell type
    nspot: int
        Number of  spots to square-bin together, along x or along y.
    remove_ambiguous_spot:
        Whether to remove ambiguous spots or not, based on pred_col
    ambiguous_spot_celltypes:
        name of ambiguous cell identify for spots, e.g. ['ambiguous']
    min_nspot_to_aggregate:
        require that at least min_nspot_to_aggregate number of spots are in the bin to get an aggregated count.
    use_csr:
        Whether to use csr format to save the binned adata.X object. Default to be True.
    """
    bins = get_bins(sthdata.adata, nspot)
    if use_csr:
        X_binned = bin_X_csr(sthdata.adata, bins)
    else:
        X_binned = bin_X(sthdata.adata, bins)
    obs_binned = bin_obs(sthdata.adata, bins)
    obsm_binned = bin_obsm(sthdata.adata, bins)
    binvar = sthdata.adata.var.copy()

    ###########  create binadata ##########
    binadata = anndata.AnnData(
        X_binned,  # raw bin count
        var=binvar,  # gene var
        obs=obs_binned,  # new bin, barcode is array_row||array_col
        obsm=obsm_binned,  # original obsm plus spatial , mean of bin barcode
        uns=sthdata.adata.uns.copy(),  # original uns, image
    )
    """
    ###########  create spatial, take mean for each bin ########## 
    spatial = pd.DataFrame(sthdata.adata.obsm['spatial'], index=sthdata.adata.obs.index, columns=['x','y'])
    for i in obs_binned.index:
        barcodes = obs_binned.loc[i]['barcodes_included']

        a = spatial.loc[barcodes]
        xc,yc = a.mean().values
        obs_binned.loc[i,'x'] = xc
        obs_binned.loc[i,'y'] = yc

    binadata.obsm['spatial'] = obs_binned[['x','y']].values
    """
    print("[Log] remove classes to be filtered out...")
    if remove_ambiguous_spot:  # Optional: remove ambiguous and filtered
        binadata = binadata[~binadata.obs[pred_col].isin(ambiguous_spot_celltypes)]
    print("[Log] filter minimum spot number...")
    binadata = binadata[binadata.obs["bin_n_spot"] >= min_nspot_to_aggregate]

    # change list to comma
    binadata.obs["barcodes_included"] = binadata.obs["barcodes_included"].str.join(
        sep=","
    )
    print(binadata)
    return binadata


def get_sthd_guided_bin_adata_v2(
    adata,
    pred_col="STHD_pred_ct",
    nspot=4,
    remove_ambiguous_spot=True,
    ambiguous_spot_celltypes=["ambiguous", "filtered"],
    min_nspot_to_aggregate=2,
    use_csr=True,
    obs_extra_col_tocollapse="",
):
    """Version 2. change input from STHD to adata.
    Bin aggregating data, guided by STHD predicted cell type identities of spots. E.g. if one bin has 2 types of spots, they will be aggregated separately, and location is determined by taking mean of x or y for the included spot barcodes
    For full side, must use use_csr. takes 1.5hr for 8million spots -> 4x4 cell type specific binning.
    #full_binadata.obs['barcodes_included'] = #full_binadata.obs['barcodes_included'].str.join(sep=',')
    #full_binadata.write('../testdata/full_STHD_guided_binadata.h5ad')
    Params
    ----------
    adata:
        adata with spatial, coodinate
    pred_col:
        column in adata.obs marking STHD predicted cell type
    nspot: int
        Number of  spots to square-bin together, along x or along y.
    remove_ambiguous_spot:
        Whether to remove ambiguous spots or not, based on pred_col
    ambiguous_spot_celltypes:
        name of ambiguous cell identify for spots, e.g. ['ambiguous']
    min_nspot_to_aggregate:
        require that at least min_nspot_to_aggregate number of spots are in the bin to get an aggregated count.
    use_csr:
        Whether to use csr format to save the binned adata.X object. Default to be True.
    obs_columns_collapse:
        Meanwhile collapsing numbers from these columns.
    """
    bins = get_bins(adata, nspot)
    if use_csr:
        X_binned = bin_X_csr(adata, bins)
    else:
        X_binned = bin_X(adata, bins)
    if obs_extra_col_tocollapse != "":
        obs_binned = bin_obs_v2(adata, bins, obs_extra_col_tocollapse)
    else:
        obs_binned = bin_obs(adata, bins)
    obsm_binned = bin_obsm(adata, bins)
    binvar = adata.var.copy()

    ###########  create binadata ##########
    binadata = anndata.AnnData(
        X_binned,  # raw bin count
        var=binvar,  # gene var
        obs=obs_binned,  # new bin, barcode is array_row||array_col
        obsm=obsm_binned,  # original obsm plus spatial , mean of bin barcode
        uns=adata.uns.copy(),  # original uns, image
    )
    """
    ###########  create spatial, take mean for each bin ########## 
    spatial = pd.DataFrame(adata.obsm['spatial'], index=adata.obs.index, columns=['x','y'])
    for i in obs_binned.index:
        barcodes = obs_binned.loc[i]['barcodes_included']

        a = spatial.loc[barcodes]
        xc,yc = a.mean().values
        obs_binned.loc[i,'x'] = xc
        obs_binned.loc[i,'y'] = yc

    binadata.obsm['spatial'] = obs_binned[['x','y']].values
    """
    print("[Log] remove classes to be filtered out...")
    if remove_ambiguous_spot:  # Optional: remove ambiguous and filtered
        binadata = binadata[~binadata.obs[pred_col].isin(ambiguous_spot_celltypes)]
    print("[Log] filter minimum spot number...")
    binadata = binadata[binadata.obs["bin_n_spot"] >= min_nspot_to_aggregate]

    # change list to comma
    binadata.obs["barcodes_included"] = binadata.obs["barcodes_included"].str.join(
        sep=","
    )
    print(binadata)
    return binadata


def main(args):
    if os.path.isfile(args.outfile):
        print("[Warning] file already exist. skipping binning...")
        sys.exit(0)
    sthdata = train.load_data_with_pdata(file_path=args.patch_path)
    if args.mode == "STHD":
        print("[Log] STHD guided binning...")
        binadata_sthd = get_sthd_guided_bin_adata(
            sthdata,
            pred_col=args.pred_col,
            nspot=args.nspot,
            remove_ambiguous_spot=args.remove_ambiguous_spot,
            ambiguous_spot_celltypes=args.ambiguous_spot_celltypes,
            min_nspot_to_aggregate=args.min_nspot_to_aggregate,
        )
        print("[Log] Saving to ", args.outfile)
        binadata_sthd.write(args.outfile)

    elif args.mode == "raw":
        print("[Log] raw binning...")
        binrawadata = get_raw_bin_adata(sthdata, nspot=args.nspot)
        print("[Log] Saving to ", args.outfile)
        binrawadata.write(args.outfile)


if __name__ == "__main__":
    """
    Example
    ----------
    python binning_fast.py --patch_path ../testdata/crop10/ --nspot 4 --outfile ../testdata/crop10_STHDbin_nspot4.h5ad
    """

    parser = argparse.ArgumentParser()
    parser.add_argument("--patch_path", type=str, help="Path of STHD result")
    parser.add_argument(
        "--nspot",
        type=int,
        default=4,
        help="number of spots to aggregate along each side",
    )
    parser.add_argument(
        "--outfile", type=str, default="../binned.h5ad", help="path to save binned h5ad"
    )
    parser.add_argument("--mode", type=str, default="STHD", help="STHD or raw")
    parser.add_argument(
        "--remove_ambiguous_spot",
        type=bool,
        default=True,
        help="Whether to remove ambiguous spots or not, based on pred_col",
    )
    parser.add_argument(
        "--ambiguous_spot_celltypes",
        type=list,
        default=["ambiguous", "filtered"],
        help="name of ambiguous cell identify for spots",
    )
    parser.add_argument(
        "--min_nspot_to_aggregate",
        type=int,
        default=2,
        help="require that at least min_nspot_to_aggregate number of spots are in the bin to get an aggregated count.",
    )
    parser.add_argument(
        "--pred_col",
        type=str,
        default="STHD_pred_ct",
        help="cell type column to aggregate",
    )
    args = parser.parse_args()

    main(args)



import copy
import json
import os

import anndata
import numpy as np
import squidpy as sq
import tifffile


class STHD:
    def __init__(self, spatial_path, counts_data, full_res_image_path, load_type):
        """Load ST-HD data.

        # When loading from original data source:
        STHD(
            spatial_path = '/hpc/group/yizhanglab/yz922/DATA/spatial/10x_HD_human_colon_cancer_20240325/square_002um/',
            counts_data = 'filtered_feature_bc_matrix.h5',
            full_res_image_path = '/hpc/group/yizhanglab/yz922/DATA/spatial/10x_HD_human_colon_cancer_20240325/Visium_HD_Human_Colon_Cancer_tissue_image.btf',
            load_type = 'original'
        )

        # When loading from a preloaded data source:
        STHD(
            spatial_path = adata,
            counts_data = None,
            full_res_image_path = '/hpc/group/yizhanglab/yz922/DATA/spatial/10x_HD_human_colon_cancer_20240325/Visium_HD_Human_Colon_Cancer_tissue_image.btf',
            load_type = 'preload'
        )

        # When loading from a cropped data source:
        STHD(
            spatial_path = '../analysis/20240331_hdcrop/crop1/adata.h5ad.gzip',
            counts_data = None,
            full_res_image_path = '../analysis/20240331_hdcrop/crop1/fullresimg_path.json',
            load_type = 'crop'
        )
        """
        if load_type == "original":
            self.adata = sq.read.visium(path=spatial_path, counts_file=counts_data)
        elif load_type == "preload":
            self.adata = spatial_path
        elif load_type == "crop":
            self.adata = anndata.read_h5ad(spatial_path)
            with open(full_res_image_path) as f:
                full_res_image_path = json.load(f)["fullresimg_path"]

        self.fullresimg_path = full_res_image_path

    def crop(self, x1, x2, y1, y2, factor):
        """Crop in original full res image's pixel coordinates.

        (x1, y1) ... (x2, y1)
        .                   .
        .                   .
        (x1, y2) ... (x2, y2)

        factor: This is the scaling factor. by default we should use the scale factor for 'hires' image.

        Example:
        -------
        x1 = 57050
        y1 = 8600
        d = 200
        x2 = x1+d
        y2 = y1+d
        #fullresimg_subset = full_data.fullresimg[y1:y2, x1:x2,:]
        #plt.imshow(fullresimg_subset)
        crop_data = full_data.crop(
            x1, x2, y1, y2,
            full_data.adata.uns['spatial']['Visium_HD_Human_Colon_Cancer']['scalefactors']['tissue_hires_scalef']
        )
        crop_data.save('../analysis/20240331_hdcrop/crop4')

        """
        # crop adata
        img = sq.im.ImageContainer.from_adata(
            self.adata
        )  # by default it got the first image, which is "hires"
        img_subset = img.crop_corner(
            int(y1 * factor),
            int(x1 * factor),
            (int((y2 - y1) * factor), int((x2 - x1) * factor)),
        )
        adata_subset = img_subset.subset(self.adata)

        # correct coordinates of the patch
        x1c, y1c, x2c, y2c = self.get_sequencing_data_region(adata_subset)
        print(f"Input Pixel is aligned to {x1c}, {y1c}, {x2c}, {y2c}")

        # crop fullresimg
        return STHD(adata_subset, None, self.fullresimg_path, "preload")

    def match_refscrna(
        self, ref, cutgene=True, gene_lambda_noise=0.000001, ref_renorm=False
    ):
        """Use reference gene expresssiong by cell type, and load to obj, adding a small noise.
        Params
        ----------
        ref: pd.DataFrame
            averaged gene expr from single cell reference, gene by celltype.

        Example:
        -------
        sec1_sthdata.adata.match_scrna_ref(genemeanpd_filtered)

        To get training dimensions:
        X = sec1_sthdata.adata.obs.shape[0] # n of spot
        Y = sec1_sthdata.adata.shape[1] # n of gene (filtered )
        Z = sec1_sthdata.lambda_cell_type_by_gene_matrix.shape[0]  # n of cell type

        """
        adata = self.adata
        adata.var_names_make_unique()
        overlap_gs = adata.var.index.intersection(ref.index)
        print(f"[Log] num of gene overlap {len( overlap_gs )}")
        if cutgene:
            adata = adata[:, adata.var_names.isin(overlap_gs)].copy()
        else:
            non_overlap_gs = [t for t in adata.var.index if t not in ref.index]
            adata = adata[:, ~non_overlap_gs]
        self.adata = adata

        ## add ref
        cell_type_by_gene_matrix = ref.loc[overlap_gs].T.values
        cell_type_by_gene_matrix = cell_type_by_gene_matrix + gene_lambda_noise
        if ref_renorm:
            print("[Log] normalizing lambda (not recommended)")
            cell_type_by_gene_matrix = (
                (cell_type_by_gene_matrix.T) / ((cell_type_by_gene_matrix.T).sum(0))
            ).T  # gene lambda for one cell type add up to 1.
        print("[Log] attaching ref gene expr in .lambda_cell_type_by_gene_matrix")
        self.lambda_cell_type_by_gene_matrix = cell_type_by_gene_matrix

    def copy(self):
        # deep copy
        new = copy.deepcopy(self)
        return new

    def load_img(self):
        img = tifffile.TiffFile(self.fullresimg_path).asarray()
        return img

    def crop_img(self, img, x1, y1, x2, y2):
        """x1 = 45450
        y1 = 4100
        d = 1100
        x2 = x1+d
        y2 = y1+d

        full_img = sthdata.load_img()
        adatacrop = sthdata.crop( x1, x2, y1, y2, factor = sthdata.adata.uns['spatial']['Visium_HD_Human_Colon_Cancer']['scalefactors']['tissue_hires_scalef'])
        x1c,x2c,y1c,y2c = adatacrop.get_sequencing_data_region()

        import matplotlib.pyplot as plt
        img_cropped = sthdata.crop_img(full_img,x1c,x2c,y1c,y2c)
        plt.imshow(img_cropped)
        """
        img_cropped = img[y1:y2, x1:x2, :]
        return img_cropped

    def get_sequencing_data_region(self, adata=None):
        if not adata:
            adata = self.adata
        x1, y1 = np.nanmin(adata.obsm["spatial"], axis=0)
        x2, y2 = np.nanmax(adata.obsm["spatial"], axis=0)
        x1, y1, x2, y2 = [int(i) for i in [x1, y1, x2, y2]]
        return x1, y1, x2, y2

    def save(self, path):
        """Save data."""
        # create path if not exist
        if not os.path.exists(path):
            print("creating new folder to save cropped data: ", path)
            os.makedirs(path)

        # save adata
        adata = self.adata
        adata.write_h5ad(os.path.join(path, "adata.h5ad.gzip"), compression="gzip")

        # save fullresimg_path
        with open(os.path.join(path, "fullresimg_path.json"), "w+") as f:
            json.dump({"fullresimg_path": self.fullresimg_path}, f, indent=4)



import os

import pandas as pd
import argparse

def format_tissue_positions(spatial_dir):
    """Transform tissue position parquet into csv. rounding to interger.
    ----------

    Example:
    -------
    format_tissue_positions(spatial_dir= '../testdata/VisiumHD/square_002um/spatial/')

    """
    tissue_parquet_path = os.path.join(spatial_dir, "tissue_positions.parquet")
    out_file = os.path.join(spatial_dir, "tissue_positions.csv")
    if os.path.isfile(out_file):
        print("File exist", out_file)
    position_parquet = pd.read_parquet(tissue_parquet_path)
    res = position_parquet.set_index("barcode").applymap(lambda t: round(t))
    res.to_csv(out_file, sep=",", header=False, index=True)
    print("Output in ", os.path.join(spatial_dir, "tissue_positions.csv"))

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--spatialdir",
        type=str,
        help="spatial directory under square_002um",
    )
    '''
    Example
    ----------
    python hdpp.py --spatialdir ../testdata/VisiumHD/binned_outputs/square_002um/spatial
    '''
    args = parser.parse_args()
    print(f'[Log] converting tissue position in {args.spatialdir}')
    format_tissue_positions(args.spatialdir)


# load reference scRNA.
# see s00_crc_singlecellreference.ipynb
import pandas as pd


def load_scrna_ref(refile):
    """
    # refile = '../testdata/crc_average_expr_genenorm_lambda_98ct_4618gs.txt'
    """
    genemeanpd_filtered = pd.read_table(refile, index_col=0)
    return genemeanpd_filtered

def gene_lambda_by_ct(adata, ctcol = 'group'):
    pdlst=[]
    #celltypes_group = celltypes.groupby(ctcol).count().index.tolist()
    celltypes_group = list(set(adata.obs[ctcol].values))
    for ct in celltypes_group:
        ctbars = adata[adata.obs[ctcol] == ct].obs.index.tolist()
        print(f'[Log] {len(ctbars)} cells for cell type {ct}')
        cellbygenecount = adata.layers['counts'][adata.obs.index.isin(ctbars), :]
        print(f'dimension of cell by gene count', cellbygenecount.shape)
        cellbygene = pd.DataFrame(cellbygenecount.todense(), index = ctbars, columns = adata.var_names)
        genebycell = cellbygene.T
        # gene normalize expr per cell (gene count ratio over all gene count), then mean over all cells from that type.
        genenorm = (genebycell/(genebycell.sum(axis=0)))
        genenorm = genenorm.mean(axis=1) # normalized gene expr for this cell type.
        pdlst.append( 
            pd.DataFrame(genenorm, columns = [ct], index = adata.var_names)
        )

    genemeanpd = pd.concat(pdlst,axis=1)
    genemeanpdfc = genemeanpd/genemeanpd.mean()
    return(genemeanpd, genemeanpdfc)

def select_ct_informative_genes_basic(genemeanpd, genemeanpdfc, exprcut = 0.000125, fccut = 2**0.5):
    mask_nomt = (~genemeanpd.index.str.startswith('MT-'))
    mask_noribo = (~genemeanpd.index.str.startswith(('RPS','RPL'))) 
    print(f'{ (mask_nomt&mask_noribo).sum() } genes after removing MT- and Ribo')
    # basic expression
    mask_expr_rctd = ( (genemeanpd>exprcut).sum(axis=1) )>0 

    print(f'{ (mask_nomt&mask_noribo&mask_expr_rctd).sum() } genes after further removing RCTD threshold gene expression 0.0125')
    ## log fold change over all cells
    mask_fc = (genemeanpdfc> fccut).sum(axis=1)>0 # rctd logfc
    print(f'{ (mask_nomt&mask_noribo&mask_expr_rctd&mask_fc).sum() } genes after further removing RCTD logfc log2 0.5')
    ## extra basic mean expr


    genemeanpd_filtered = genemeanpd[mask_nomt&mask_noribo&mask_expr_rctd&mask_fc]
    print(genemeanpd_filtered.shape)
    return(genemeanpd_filtered)
